#!/usr/bin/env node
'use strict';

var path = require('path');
var chalk = require('chalk');
var osLocale = require('os-locale');
var Compiler = require('../compiler');
var MessageWriter = require('../libs/message-writer.js');

/**
 * uoTES.js コンパイラモジュール
 * コンパイルコマンド　メインクラス
 * 
 * @since 2015/12/16
 * @author RaTTiE
 */
class Main {
	/**
	 * コンストラクタ。
	 */
	constructor() {
		this.commander = require('commander');
		this.compiler = new Compiler();
	}

	/**
	 * 処理を開始する。
	 */
	start() {
		if (!this.setup()) return;
		if (!this.compile()) return;
	}

	/**
	 * セットアップを行う。
	 */
	setup() {
		if (!this.setupCommander()) return false;
		if (!this.setupResources()) return false;
		if (!this.setupEngine()) return false;
		if (!this.setupDirs()) return false;

		return true;
	}

	/**
	 * Commanderのセットアップを行う。
	 */
	setupCommander() {
		this.commander
			.version(this.compiler.version)
			.usage('[options] <file ...>')
			.option('-m, --modules <modules>', 'assign modules', String)
			.option('-r, --replacer <replacer>', 'assign replacer file', String)
			.option('-i, --in <indir>', 'set input directory', String)
			.option('-o, --out <outdir>', 'set output directory', String)
			.parse(process.argv);

		return true;
	}

	/**
	 * アプリケーションリソースのセットアップ処理を行う。
	 */
	setupResources() {
		this.compiler.addResourcePath(path.resolve(__dirname, '../resources'));
		this.compiler.language = osLocale.sync();

		this.compiler.loadStringResource();

		return true;
	}

	/**
	 * エンジンのセットアップを行う。
	 */
	setupEngine() {
		var modules = ['mv'];

		if (this.commander.modules) {
			modules = this.commander.modules.split(',');
		}

		try {
			this.compiler.loadModules(modules);
		} catch (err) {
			console.error(chalk.bold.red(err.stack));

			return false;
		}
		this.compiler.messageWriter = new MessageWriter(this.compiler.stringResource);

		if (this.commander.replacer) {
			try {
				if (!this.compiler.loadReplacer(this.commander.replacer)) {
					return false;
				}
			} catch (err) {
				console.error(chalk.bold.red(err.stack));

				return false;
			}
		}

		return true;
	}

	/**
	 * ディレクトリのセットアップを行う。
	 */
	setupDirs() {
		if (this.commander.in) this.compiler.inDir = this.commander.in;
		if (this.commander.out) this.compiler.outDir = this.commander.out;

		return true;
	}

	/**
	 * ソースコードのコンパイルを行う。
	 */
	compile() {
		var files = this.commander.args;
		if (files.length < 1) {
			this.commander.help();

			return false;
		}

		try {
			return this.compiler.execute(files);
		} catch (err) {
			console.error(chalk.bold.red(err.stack));

			return false;
		}
	}
}

var main = new Main();

main.start();
