#!/usr/bin/env node
'use strict';

var path = require('path');
var chalk = require('chalk');
var osLocale = require('os-locale');
var Decompiler = require('../decompiler');
var MessageWriter = require('../libs/message-writer.js');

/**
 * uoTES.js コンパイラモジュール
 * デコンパイルコマンド　メインクラス
 * 
 * @since 2016/03/04
 * @author RaTTiE
 */
class Main {
	/**
	 * コンストラクタ。
	 */
	constructor() {
		this.commander = require('commander');
		this.decompiler = new Decompiler();
	}

	/**
	 * 処理を開始する。
	 */
	start() {
		if (!this.setup()) return;
		if (!this.decompile()) return;
	}

	/**
	 * セットアップを行う。
	 */
	setup() {
		if (!this.setupCommander()) return false;
		if (!this.setupResources()) return false;
		if (!this.setupEngine()) return false;
		if (!this.setupDirs()) return false;

		return true;
	}

	/**
	 * Commanderのセットアップを行う。
	 */
	setupCommander() {
		var packageJson = require('../package.json');

		this.commander
			.version(this.decompiler.version)
			.usage('[options] <file ...>')
			.option('-m, --modules <modules>', 'assign modules', String)
			.option('-i, --in <indir>', 'set input directory', String)
			.option('-o, --out <outdir>', 'set output directory', String)
			.option('-e, --extention <extention>', 'set output script file extention', String)
			.parse(process.argv);

		return true;
	}

	/**
	 * アプリケーションリソースのセットアップ処理を行う。
	 */
	setupResources() {
		this.decompiler.addResourcePath(path.resolve(__dirname, '../resources'));
		this.decompiler.language = osLocale.sync();

		this.decompiler.loadStringResource();

		return true;
	}

	/**
	 * エンジンのセットアップを行う。
	 */
	setupEngine() {
		var modules = ['mv'];

		if (this.commander.modules) {
			modules = this.commander.modules.split(',');
		}

		if (this.commander.extention) {
			this.decompiler.outputExtention = this.commander.extention;
		}

		try {
			this.decompiler.loadModules(modules);
		} catch (err) {
			console.error(chalk.bold.red(err.stack));

			return false;
		}
		this.decompiler.messageWriter = new MessageWriter(this.decompiler.stringResource);

		return true;
	}

	/**
	 * ディレクトリのセットアップを行う。
	 */
	setupDirs() {
		if (this.commander.in) this.decompiler.inDir = this.commander.in;
		if (this.commander.out) this.decompiler.outDir = this.commander.out;

		return true;
	}

	/**
	 * イベントデータのデコンパイルを行う。
	 */
	decompile() {
		var files = this.commander.args;
		if (files.length < 1) {
			this.commander.help();

			return false;
		}

		try {
			return this.decompiler.execute(files);
		} catch (err) {
			console.error(chalk.bold.red(err.stack));

			return false;
		}
	}
}

var main = new Main();

main.start();
