'use strict';

var fs = require('fs-extra');
var path = require('path');
var Translator = require('./libs/translator');

/**
 * uoTES.js コンパイラモジュール
 * コンパイラクラス
 * 
 * @since 2016/04/06
 * @author RaTTiE
 */
class Compiler extends Translator {
	/**
	 * コンストラクタ。
	 */
	constructor() {
		super();
	}

	/**
	 * コンパイルを行う。
	 * 
	 * @param files コンパイル対象ファイルの一覧。
	 * @return 処理成功時はtrue。エラー時はfalse。
	 */
	execute(files) {
		if (!this.verify()) return false;

		var hasError = false;
		for (var i = 0; i < files.length; i++) {
			var inFile = files[i];
			try {
				var inPath = path.resolve(this.inDir, inFile);

				var source;
				try {
					source = fs.readFileSync(inPath, 'utf8');
				} catch (err) {
					if (err.code == 'ENOENT') {
						this.console.error(this.stringResource.get('tes.scenario.notFound.error'));

						hasError = true;
						continue;
					} else {
						this.console.error(this.stringResource.get('tes.scenario.load.failure'));

						throw err;
					}
				}

				this.engine.messageWriter.fileName = inFile;
				var result = this.engine.compile(source);

				if (result.error != null) {
					hasError = true;
					continue;
				}
			} catch (err) {
				this.console.error(this.stringResource.get('tes.compile.failure'));

				throw err;
			}

			var outDir = path.resolve(this.outDir, path.dirname(inFile));
			if (!fs.existsSync(outDir)) {
				this.console.log(this.stringResource.get('dir.creating', outDir));

				try {
					fs.mkdirsSync(outDir);
				} catch (err) {
					this.console.error(this.stringResource.get('dir.creating.error'));

					throw err;
				}
			}

			try {
				var outFile = path.basename(inFile, path.extname(inFile)) + '.json';
				var outPath = path.resolve(outDir, outFile);

				this.console.log(this.stringResource.get('tes.event.writing', outFile));
				fs.writeFileSync(outPath, JSON.stringify(result.list));
			} catch (err) {
				this.console.error(this.stringResource.get('tes.event.write.failure'));

				throw err;
			}
		}

		if (hasError) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * 変換辞書のロードを行う。
	 */
	loadReplacer(filePath) {
		try {
			var extname = path.extname(filePath);
			this.console.log(this.stringResource.get('replacer.loading', filePath));
			var source = fs.readFileSync(filePath, 'utf8');

			if (extname == '.json') {
				this.engine.addReplacersFromObject(JSON.parse(source)); 
			} else {
				this.engine.addReplacersFromText(source);
			}

			return true;
		} catch (err) {
			if (err.code == 'ENOENT') {
				this.console.error(this.stringResource.get('replacer.notFound.error'));

				return false;
			} else {
				this.console.error(this.stringResource.get('replacer.load.failure'));

				throw err;
			}
		}
	}
}

module.exports = Compiler;
