'use strict';

var fs = require('fs-extra');
var path = require('path');
var Translator = require('./libs/translator');

/**
 * uoTES.js コンパイラモジュール
 * デコンパイラクラス
 * 
 * @since 2016/04/07
 * @author RaTTiE
 */
class Decompiler extends Translator {
	/**
	 * コンストラクタ。
	 */
	constructor() {
		super();

		this.outputExtention = 'txt';
	}

	/**
	 * デコンパイルを行う。
	 * 
	 * @param files コンパイル対象ファイルの一覧。
	 * @return 処理成功時はtrue。エラー時はfalse。
	 */
	execute(files) {
		if (!this.verify()) return false;

		var hasError = false;
		var result;
		for (var i = 0; i < files.length; i++) {
			var inFile = files[i];
			var inPath = path.resolve(this.inDir, inFile);
			var basename = path.basename(inFile, path.extname(inFile));

			try {
				var data;
				try  {
					data = JSON.parse(fs.readFileSync(inPath, 'utf8'));
				} catch (err) {
					if (err.code == 'ENOENT') {
						this.console.error(this.stringResource.get('tes.event.notFound.error'));

						hasError = true;
						continue;
					} else {
						this.console.error(this.stringResource.get('tes.event.load.failure'));

						throw err;
					}
				}

				this.engine.messageWriter.fileName = inFile;
				result = this.engine.decompile(basename, data);

				if (result == null) {
					hasError = true;
					continue;
				}
			} catch (err) {
				this.console.error(this.stringResource.get('tes.decompile.failure'));

				throw err;
			}

			var outDir = path.resolve(this.outDir, path.dirname(inFile));
			if (!fs.existsSync(outDir)) {
				this.console.log(this.stringResource.get('dir.creating', outDir));

				try {
					fs.mkdirsSync(outDir);
				} catch (err) {
					this.console.error(this.stringResource.get('dir.creating.error'));

					throw err;
				}
			}

			try {
				var outFile;
				var outPath;
				var linkData = {
					type: result.type,
					targets: []
				};

				for (var j = 0; j < result.targets.length; j++) {
					var target = result.targets[j];

					if (target.error != null) {
						hasError = true;
						continue;
					}

					outFile = target.name + '.' + this.outputExtention;
					outPath = path.resolve(outDir, outFile);

					this.console.log(this.stringResource.get('tes.scenario.writing', outFile));
					fs.writeFileSync(outPath, target.source);

					delete target.list;
					delete target.source;

					linkData.targets.push(target);
				}
			} catch (err) {
				this.console.error(this.stringResource.get('tes.scenario.write.failure'));

				throw err;
			}

			if (linkData.type != 'tes-event') {
				try {
					outFile = basename + '_link.json';
					outPath = path.resolve(outDir, outFile);

					this.console.log(this.stringResource.get('tes.link.writing', outFile));
					fs.writeFileSync(outPath, JSON.stringify(linkData));
				} catch (err) {
					this.console.error(this.stringResource.get('tes.link.write.failure'));

					throw err;
				}
			}
		}

		if (hasError) {
			return false;
		} else {
			return true;
		}
	}
}

module.exports = Decompiler;
