'use strict';

var fs = require('fs-extra');
var path = require('path');
var Tes = require('uotes-core');
var StringResourceManager = require('es-common-core/string-resource-manager');

/**
 * uoTES.js コンパイラモジュール
 * トランスレータクラス
 * 
 * @since 2016/04/07
 * @author RaTTiE
 */
class Translator {
	/**
	 * コンストラクタ。
	 */
	constructor() {
		this.version = require('../package.json').version;
		this.stringResource = new StringResourceManager();
		this.engine = new Tes.Engine();
		this.console = console;
		this.inDir = './';
		this.outDir = './';
	}

	/**
	 * モジュールのルートパスを取得する。
	 */
	get moduleRootPath() {
		return path.resolve(__dirname, '..');
	}

	/**
	 * リソースパスを追加する。
	 */
	addResourcePath(resourcePath) {
		this.stringResource.addPath(resourcePath);
	}

	/** 言語を取得する。 */
	get language() {
		return this.stringResource.language;
	}

	/** 言語を設定する。 */
	set language(value) {
		this.stringResource.language = value;
	}

	/** メッセージライターを取得する。 */
	get messageWriter() {
		return this.engine.messageWriter;
	}

	/** メッセージライターを設定する。 */
	set messageWriter(value) {
		this.engine.messageWriter = value;
	}

	/**
	 * 文字列リソースのロードを行う。
	 */
	loadStringResource() {
		var that = this;
		var existsLanguage = this.stringResource.list().find((element, index, array) => {
			return element.language == that.stringResource.language;
		}) != null;
		if (!existsLanguage) this.stringResource.language = 'default';
		
		this.stringResource.loadFromObject(Tes.StringResourceData);
		this.stringResource.load(this.stringResource.language, true);
	}

	/**
	 * モジュールのロードを行う。
	 * 
	 * @param moduleNames ロード対象モジュール名の一覧。
	 */
	loadModules(moduleNames, showMessage) {
		if (showMessage == null) showMessage = true;

		for (var i = 0; i < moduleNames.length; i++) {
			this.loadModule(moduleNames[i], showMessage);
		}
	}

	/**
	 * モジュールのロードを行う。
	 * 
	 * @param moduleName ロード対象モジュール名。
	 */
	loadModule(moduleName, showMessage) {
		var module = require('uotes-module-' + moduleName);

		try {
			if (showMessage) {
				this.console.log(this.stringResource.get('module.loading', moduleName));
			}
			this.engine.addModule(module);
			if (module.stringResourceData) {
				this.stringResource.loadFromObject(
					module.stringResourceData,
					this.stringResource.language,
					true
				);
			}
		} catch (err) {
			this.console.error(this.stringResource.get('module.load.failure', moduleName));
  
			throw err;
		}
	}

	/**
	 * ディレクトリのセットアップを行う。
	 * 
	 * @return 処理成功時はtrue。エラー時はfalse。
	 */
	setupDirs() {
		this.inDir = path.resolve(this.inDir);
		this.outDir = path.resolve(this.outDir);

		if (fs.existsSync(this.inDir)) {
			var stat = fs.statSync(this.inDir);

			if (stat.isFile()) {
				this.console.error(this.stringResource.get('dir.notdir.error', this.inDir));

				return false;
			}
		} else {
			this.console.error(this.stringResource.get('dir.notFound.error', this.inDir));

			return false;
		}

		if (fs.existsSync(this.outDir)) {
			var stat = fs.statSync(this.outDir);
			
			if (stat.isFile()) {
				this.console.error(this.stringResource.get('dir.notdir.error', this.outDir));

				return false;
			}
		} else {
			this.console.log(this.stringResource.get('dir.creating', this.outDir));

			try {
				fs.mkdirsSync(this.outDir);
			} catch (err) {
				this.console.error(this.stringResource.get('dir.creating.error'));

				throw err;
			}
		}

		return true;
	}

	/**
	 * 変換処理を行う。
	 * 
	 * @return 処理成功時はtrue。エラー時はfalse。
	 */
	verify() {
		if (!this.setupDirs()) return false;

		return true;
	}

	/**
	 * 変換処理を行う。
	 * 
	 * @param files 変換処理対象ファイルの一覧。
	 * @return 処理成功時はtrue。エラー時はfalse。
	 */
	execute(files) {
		throw new Error('Please implement Translator.execute');
	}
}

module.exports = Translator;
