(function(definition) {
'use strict';

if (typeof exports === 'object') {
	// CommonJS
	module.exports = definition();
} else {
	// <script>
	window.Tes = definition();
}
})(function () {
'use strict';

var Tes = {};

/**
 * エンジンクラス
 */
var TesEngine = function() {
	this.initialize.apply(this, arguments);
};
TesEngine.prototype = Object.create(Object.prototype);
TesEngine.prototype.constructor = TesEngine;

/**
 * 初期化
 */
TesEngine.prototype.initialize = function() {
	this.loader = new TesLoader(this);
	this.splitter = new TesSplitter(this);
	this.validator = new TesValidator(this);
	this.converter = new TesConverter(this);
	this.deconverter = new TesDeconverter(this);
	this.linker = new TesLinker(this);
	this.messageWriter = new TesDefaultMessageWriter();
};

/**
 * コンパイルを実行する。
 */
TesEngine.prototype.compile = function(source) {
	var target = {
		source: source
	};

	this.loader.loadFromText(source);
	if (!this.loader.isError()) {
		target.loadedData = this.loader.getData();
	} else {
		target.error = this.loader.getError();

		return target;
	};

	this.splitter.run(this.loader.getData());
	if (!this.splitter.isError()) {
		target.splitedData = this.splitter.getData();
	} else {
		target.error = this.splitter.getError();

		return target;
	};

	this.validator.run(this.splitter.getData());
	if (this.validator.isError()) {
		target.error = this.validator.getError();
		
		return target;
	}

	this.converter.run(this.validator.getData());
	if (!this.converter.isError()) {
		target.list = this.converter.getData();
	} else {
		target.error = this.converter.getError();

		return target;
	};

	return target;
}

/**
 * デコンパイルを実行する。
 */
TesEngine.prototype.decompile = function(basename, data) {
	try {
		var linkData = this.deconverter.identifyTargets(basename, data);
	} catch (err) {
		if (err.name == 'DeconvertError') {
			this.writeError(err);

			return null;
		} else {
			throw err;
		}
	}

	for (var i = 0; i < linkData.targets.length; i++) {
		var target = linkData.targets[i];

		this.deconverter.run(target);
		if (!this.deconverter.isError()) {
			target.source = this.deconverter.getSource();
		} else {
			target.error = this.deconverter.getError();
		}
	}

	return linkData;
};

/**
 * リンク処理を実行する。
 */
TesEngine.prototype.link = function(linkType, lineNumber, target, targetData, eventList) {
	var source = {
		linkType: linkType,
		target: target,
		targetData: targetData,
		eventList: eventList,
		lineNumber: lineNumber
	};

	this.linker.run(source);
	if (this.linker.isError()) {
		source.error = this.linker.getError();
	}

	return source;
}

/**
 * 変換辞書をクリアする
 */
TesEngine.prototype.clearReplacers = function() {
	this.splitter._replacers = {};
}

/**
 * 変換辞書をセットする
 */
TesEngine.prototype.addReplacersFromObject = function(replacers) {
	for (var key in replacers) {
		var value = replacers[key];

		if (!RubyCompatibility.isEmptyText(key) && !RubyCompatibility.isEmptyText(value)) {
			key = RubyCompatibility.stripText(key);
			value = RubyCompatibility.stripText(value);

			this.splitter._replacers[key] = value;
		}
	}
};

/**
 * 変換辞書を文字列から作成
 */
TesEngine.prototype.addReplacersFromText = function(text) {
	text = text.replace(/\r\n|\n/g, '\r');

	var lines = text.split('\r');
	for (var i = 0; i < lines.length; i++) {
		var line = lines[i];

		line = line.replace(/\r\n|\r|\n/g, '');

		var splited = line.split(/\t+/);

		if (!RubyCompatibility.isEmptyText(splited[0]) && !RubyCompatibility.isEmptyText(splited[1])) {
			splited[0] = RubyCompatibility.stripText(splited[0]);
			splited[1] = RubyCompatibility.stripText(splited[1]);

			this.splitter._replacers[splited[0]] = splited[1];
		}
	}
};

/**
 * 分割関数をセット
 */
TesEngine.prototype.addSplitFuncs = function(splitFuncs) {
	for (var key in splitFuncs) {
		this.splitter._splitFuncs[key] = splitFuncs[key];
	}
};

/**
 * 分割関数をクリア
 */
TesEngine.prototype.clearSplitFuncs = function() {
	this.splitter._splitFuncs = {};
};

/**
 * メッセージ行数をセット
 */
TesEngine.prototype.setMessageLineSize = function(messageLineSize) {
	this.splitter._messageLineSize = messageLineSize;
}

/**
 * コマンド検証設定をセット
 */
TesEngine.prototype.addValidates = function(validates) {
	for (var key in validates) {
		this.validator._validates[key] = validates[key];
	}
};

/**
 * コマンド検証設定をクリア
 */
TesEngine.prototype.clearValidates = function() {
	this.validator._validates = {};
};

/**
 * コマンド検証関数をセット
 */
TesEngine.prototype.addValidateFuncs = function(validateFuncs) {
	this.validator.addValidateFuncs(validateFuncs);
};

/**
 * コマンド検証関数をクリア
 */
TesEngine.prototype.clearValidateFuncs = function() {
	this.validator._validateFuncs = {};
};

/**
 * イベントデータ変換関数をセット
 */
TesEngine.prototype.addConvertFuncs = function(convertFuncs) {
	for (var key in convertFuncs) {
		this.converter._convertFuncs[key] = convertFuncs[key];
	}
};

/**
 * イベントデータ変換関数をクリア
 */
TesEngine.prototype.clearConvertFuncs = function() {
	this.converter._convertFuncs = {};
};

/**
 * イベントデータ逆変換関数をセット
 */
TesEngine.prototype.addDeconvertFuncs = function(deconvertFuncs) {
	for (var key in deconvertFuncs) {
		this.deconverter._deconvertFuncs[key] = deconvertFuncs[key];
	}
};

/**
 * イベントデータ逆変換関数をクリア
 */
TesEngine.prototype.clearDeconvertFuncs = function() {
	this.deconverter._deconvertFuncs = {};
};

/**
 * デコンバート対象判定関数を追加
 */
TesEngine.prototype.addDeconvertTargetIdetifier = function(targetIdentifier) {
	this.deconverter._targetIdentifiers.push(targetIdentifier);
};

/**
 * デコンバート対象判定関数をクリア
 */
TesEngine.prototype.clearDeconvertTargetIentifier = function() {
	this.deconverter._targetIdentifiers = [];
};

/**
 * リンク関数を追加
 */
TesEngine.prototype.addLinkFuncs = function(linkFuncs) {
	for (var key in linkFuncs) {
		this.linker._linkFuncs[key] = linkFuncs[key];
	}
};

/**
 * リンク関数をクリア
 */
TesEngine.prototype.clearLinkFuncs = function() {
	this.linker._linkFuncs = {};
}

/**
 * リンク関数の存在チェックを行う
 */
TesEngine.prototype.existsLinkFunction = function(linkType) {
	return this.linker._linkFuncs[linkType] != null;
};

/**
 * オーバーライド処理をセット
 */
TesEngine.prototype.addProcessors = function(module) {
	if (module.preProcessors) {
		for (var key in module.preProcessors) {
			if (this[key] != null) {
				this[key]._preProcessors.push(module.preProcessors[key]);
			}
		}
	}

	if (module.processors) {
		for (var key in module.processors) {
			if (this[key] != null) {
				this[key]._processors.push(module.processors[key]);
			}
		}
	}

	if (module.postProcessors) {
		for (var key in module.postProcessors) {
			if (this[key] != null) {
				this[key]._postProcessors.push(module.postProcessors[key]);
			}
		}
	}
};

/**
 * オーバーライド処理をクリア
 */
TesEngine.prototype.clearProcessors = function() {
	for (var key in ['loader', 'splitter', 'validator', 'converter', 'deconverter', 'link']) {
		this['key'].clearProcessors();
	}
}

/**
 * 文法設定データをセット
 */
TesEngine.prototype.addModule = function(module) {
	if (module == null) return;

	if (module.replacers != null) this.addReplacersFromObject(module.replacers);
	if (module.splitFuncs != null) this.addSplitFuncs(module.splitFuncs);
	if (module.validates != null) this.addValidates(module.validates);
	if (module.validateFuncs != null) this.addValidateFuncs(module.validateFuncs);
	if (module.convertFuncs != null) this.addConvertFuncs(module.convertFuncs);
	if (module.deconvertFuncs != null) this.addDeconvertFuncs(module.deconvertFuncs);
	if (module.linkFuncs != null) this.addLinkFuncs(module.linkFuncs);
	if (module.decompileTargetIdentifier != null) this.addDeconvertTargetIdetifier(module.decompileTargetIdentifier);
	if (module.messageLineSize != null) this.setMessageLineSize(module.messageLineSize);
	this.addProcessors(module);
};

/**
 * 文法設定データをクリア
 */
TesEngine.prototype.clearModule = function() {
	this.clearReplacers();
	this.clearSplitFuncs();
	this.clearValidates();
	this.clearValidateFuncs();
	this.clearConvertFuncs();
	this.clearDeconvertFuncs();
	this.clearLinkFuncs();
	this.clearDeconvertTargetIentifier();
	this.setMessageLineSize(4);
	this.clearProcessors();
};

/**
 * メッセージを出力する
 */
TesEngine.prototype.writeMessage = function(message, code, args, address) {
	this.messageWriter.writeMessage.call(this.messageWriter, message, code, args, address);
}

/**
 * エラーメッセージを出力する
 */
TesEngine.prototype.writeErrorMessage = function(message, code, args, address) {
	this.messageWriter.writeMessage.call(this.messageWriter, message, code, args, address);
}

/**
 * 警告メッセージを出力する
 */
TesEngine.prototype.writeWarningMessage = function(message, code, args, address) {
	this.messageWriter.writeMessage.call(this.messageWriter, message, code, args, address);
}

/**
 * エラーを出力する
 */
TesEngine.prototype.writeError = function(err, address) {
	this.messageWriter.writeError.call(this.messageWriter, err, address);
}

/**
 * パラメータの区切り記号をセットする
 */
TesEngine.prototype.setDeconvertQuote = function(quote) {
	this.deconverter.quote = quote;
}
Tes.Engine = TesEngine;

/**
 * デフォルトメッセージライタークラス
 */
var TesDefaultMessageWriter = function() {
	this.initialize.apply(this, arguments);
};
TesDefaultMessageWriter.prototype = Object.create(Object.prototype);
TesDefaultMessageWriter.prototype.constructor = TesDefaultMessageWriter;

/**
 * 初期化
 */
TesDefaultMessageWriter.prototype.initialize = function() {};

/**
 * メッセージを出力する
 */
TesDefaultMessageWriter.prototype.writeMessage = function(message, code, args, address) {
	console.log(message);
};

/**
 * エラーメッセージを出力する
 */
TesDefaultMessageWriter.prototype.writeErrorMessage = function(message, code, args, address) {
	console.error(message);
};

/**
 * 警告メッセージを出力する
 */
TesDefaultMessageWriter.prototype.writeWarningMessage = function(message, code, args, address) {
	console.warn(message);
};

/**
 * エラーを出力する
 */
TesDefaultMessageWriter.prototype.writeError = function(err, address) {
	console.error(err.message);
};

/**
 * ロード時エラー
 */
var TesLoadError = function(message, code, args) {
	this.message = message;
	this.code = code;
	this.args = args;
};
TesLoadError.prototype = new Error();
TesLoadError.prototype.constructor = TesLoadError;
TesLoadError.prototype.name = 'LoadError';
Tes.LoadError = TesLoadError;

/**
 * 文法エラー
 */
var TesSyntaxError = function(message, code, args) {
	this.message = message;
	this.code = code;
	this.args = args;
};
TesSyntaxError.prototype = new Error();
TesSyntaxError.prototype.constructor = TesSyntaxError;
TesSyntaxError.prototype.name = 'SyntaxError';
Tes.SyntaxError = TesSyntaxError;

/**
 *コマンド分割エラー
 */
var TesSplitError = function(message, code, args) {
	this.message = message;
	this.code = code;
	this.args = args;
};
TesSplitError.prototype = new Error();
TesSplitError.prototype.constructor = TesSplitError;
TesSplitError.prototype.name = 'SplitError';
Tes.SplitError = TesSplitError;

/**
 * 検証エラー
 */
var TesValidateError = function(message, code, args) {
	this.message = message;
	this.code = code;
	this.args = args;
};
TesValidateError.prototype = new Error();
TesValidateError.prototype.constructor = TesValidateError;
TesValidateError.prototype.name = 'ValidateError';
Tes.ValidateError = TesValidateError;

/**
 * 変換失敗エラー
 */
var TesConvertError = function(message, code, args) {
	this.message = message;
	this.code = code;
	this.args = args;
};
TesConvertError.prototype = new Error();
TesConvertError.prototype.constructor = TesConvertError;
TesConvertError.prototype.name = 'ConvertError';
Tes.ConvertError = TesConvertError;

/**
 * 逆変換失敗エラー
 */
var TesDeconvertError = function(message, code, args) {
	this.message = message;
	this.code = code;
	this.args = args;
};
TesDeconvertError.prototype = new Error();
TesDeconvertError.prototype.constructor = TesDeconvertError;
TesDeconvertError.prototype.name = 'DeconvertError';
Tes.DeconvertError = TesDeconvertError;

/**
 * リンク失敗エラー
 */
var TesLinkError = function(message, code, args) {
	this.message = message;
	this.code = code;
	this.args = args;
};
TesLinkError.prototype = new Error();
TesLinkError.prototype.constructor = TesLinkError;
TesLinkError.prototype.name = 'LinkError';
Tes.LinkError = TesLinkError;

/**
 * TES変換ベースクラス
 */
var TesConverterBase = function(engine) {
	this.initialize.apply(this, arguments);

	this.clearProcessors();
};
TesConverterBase.prototype = Object.create(Object.prototype);
TesConverterBase.prototype.constructor = TesConverterBase;

/**
 * 初期化
 */
TesConverterBase.prototype.initialize = function(engine) {
	this.engine = engine;
	this._data = [];
	this.reset();
	this.clearProcessors();
};

/**
 * 処理開始
 */
TesConverterBase.prototype.run = function() {
	this.reset();

	this.preProcessing();
	for (var i = 0; i < this._preProcessors.length; i++) {
		this._preProcessors[i].call(this);
	}

	this.processing();
	for (var i = 0; i < this._processors.length; i++) {
		this._processors[i].call(this);
	}

	this.postProcessing();
	for (var i = 0; i < this._postProcessors.length; i++) {
		this._postProcessors[i].call(this);
	}
};

/**
 * リセット処理
 */
TesConverterBase.prototype.reset = function() {
	this._error = null;
};

/**
 * 前処理
 */
TesConverterBase.prototype.preProcessing = function() {};

/**
 * メイン処理
 */
TesConverterBase.prototype.processing = function() {};

/**
 * 後処理
 */
TesConverterBase.prototype.postProcessing = function() {
	if (this.isError()) {
		this._data = [];
	}
};

/** データ取得 */
TesConverterBase.prototype.getData = function() {
	return this._data;
};

// エラー取得
TesConverterBase.prototype.getError = function() {
	return this._error;
}

/** エラーフラグ */
TesConverterBase.prototype.isError = function() {
	if (this._data.length == 0) return true;

	return this._error != null;
};

/** 行数 */
TesConverterBase.prototype.getLineNumber = function() {
	return 0;
};

/** エラー出力 */
TesConverterBase.prototype.writeError = function(err) {
	this._error = err;

	this.engine.writeError(err, this.getLineNumber());
};

/**
 * 空白文字を取り除く
 */
TesConverterBase.prototype.stripText = function(text) {
	return RubyCompatibility.stripText(text);
}

/**
 * 空白文字か？
 */
TesConverterBase.prototype.isEmptyText = function(text) {
	return RubyCompatibility.isEmptyText(text);
}

/**
 * オーバーライド処理をクリアする
 */
TesConverterBase.prototype.clearProcessors = function() {
	this._preProcessors = [];
	this._processors = [];
	this._postProcessors = [];
}

/**
 * ＴＥＳローダクラス
 */
var TesLoader = function(engine) {
	this.initialize.apply(this, arguments);
};

TesLoader.prototype = Object.create(TesConverterBase.prototype);
TesLoader.prototype.constructor = TesLoader;

/**
 * 初期化
 */
TesLoader.prototype.initialize = function(engine) {
	TesConverterBase.prototype.initialize.call(this, engine);
};

/**
 * 文字列からロードを行う
 */
TesLoader.prototype.loadFromText = function(text) {
	this._text = text;
	this._data = [];

	this.run();
};

/**
 * リセット処理
 */
TesLoader.prototype.reset = function() {
	TesConverterBase.prototype.reset.call(this);

	this._lineNumber = 0;
};

/**
 * 処理開始
 */
TesLoader.prototype.run = function() {
	this.engine.writeMessage('>TESの読み込みを開始', 'loader.start');

	try {
		TesConverterBase.prototype.run.call(this);
	} catch (err) {
		this.writeError(err);
		this.engine.writeErrorMessage('>>TESの読み込みに失敗', 'loader.failure');
	}

	this.engine.writeMessage('>>>TESの読み込み処理が終了\n', 'loader.complete');
};

/**
 * メイン処理
 */
TesLoader.prototype.processing = function() {
	TesConverterBase.prototype.processing.call(this);

	if (this._text != null) {
		this._text = this._text.replace(/\r\n|\n/g, '\r');

		var lines = this._text.split('\r');
		for (var i = 0; i < lines.length; i++) {
			this._data.push({
				line: lines[i],
				lineNumber: ++this._lineNumber
			});
		}
	}
};

/**
 * 後処理
 */
TesLoader.prototype.postProcessing = function() {
	TesConverterBase.prototype.postProcessing.call(this);

	if (!this.isError() && this._data.length == 0) {
		throw new TesLoadError('\n>>変換できるシナリオがありませんでした', 'loader.nodata.error');
	}
};

/** 行数 */
TesLoader.prototype.getLineNumber = function() {
	return this._lineNumber;
};

/**
 * ＴＥＳ分割クラス
 */
var TesSplitter = function(engine) {
	this.initialize.apply(this, arguments);
};
TesSplitter.prototype = Object.create(TesConverterBase.prototype);
TesSplitter.prototype.constructor = TesConverter;

/**
 * 初期化
 */
TesSplitter.prototype.initialize = function(engine) {
	TesConverterBase.prototype.initialize.call(this, engine);

	this._replacers = {};
	this._splitFuncs = {};
	this._messageLineSize = 4;
};

/**
 * リセット処理
 */
TesSplitter.prototype.reset = function() {
	TesConverterBase.prototype.reset.call(this);

	this._index = 0;
};

/**
 * 処理開始
 */
TesSplitter.prototype.run = function(lines) {
	this._lines = lines;
	this._data = [];

	this.engine.writeMessage('>TESのコマンド分割を開始', 'splitter.start');

	try {
		TesConverterBase.prototype.run.call(this);
	} catch (err) {
		this.writeError(err);
		this.engine.writeErrorMessage('>>TESのコマンド分割に失敗', 'splitter.failure');
	}

	this.engine.writeMessage('>>>TESのコマンド分割処理が完了\n', 'splitter.complete');
};

/**
 * メイン処理
 */
TesSplitter.prototype.processing = function() {
	TesConverterBase.prototype.processing.call(this);

	while (!this.eoc()) {
		try {
			this.allocate();
		} catch (err) {
			if (err.name == 'SyntaxError') {
				this.writeError(err);
			} else if (err.name == 'SplitError') {
				this.writeError(err);
			} else {
				throw err;
			}
		} finally {
			this._index++;
		}
	}
};

/**
 * 後処理
 */
TesSplitter.prototype.postProcessing = function() {
	TesConverterBase.prototype.postProcessing.call(this);

	if (!this.isError() && this._data.length == 0) {
		throw new TesSplitError(
			'\n>>TESコマンドが検出されませんでした',
			'splitter.nodata.error'
		);
	}
};

/**
 * 分割処理の振り分け
 */
TesSplitter.prototype.allocate = function() {
	var text = this.getLine();

	if (RubyCompatibility.isEmptyText(text)) return;

	var first = this.getFirstCharacter(text);

	switch (first) {
		case '*':
			// ラベル
			this.splitLabel(text);

			break;
		case '@':
			// タグ
			var tag = this.getCommandName(text);
			if (this._splitFuncs[tag]) {
				// 専用分割処理
				this._splitFuncs[tag].call(this, text, tag);
			} else {
				// 通常分割処理
				this.splitNormal(text, tag);
			}

			break;
		default:
			// メッセージ行分割
			this.splitMessage(text);

			break;
	}
};

/**
 * 分割処理（メッセージ）
 */
TesSplitter.prototype.splitMessage = function(text) {
	if (this._data.length == 0 || this._data[this._data.length - 1].tag != 'message_h') {
		this.addData('message_h');
	}

	var messages = [];
	while (!this.eoc()) {
		// アンダースコアの置換を実行
		text = this.replaceMessageUnderscore(text);

		messages.push(
			this.createData(
				'message', {
					value: this.replaceMessage(text)
				}
			)
		);

		if (messages.length >= this.getMessageLineSize() || !this.nextMessage()) {
			break;
		}

		this._index++;
		text = this.getLine();
	}

	for (var i = 0; i < messages.length; i++) {
		this._data.push(messages[i]);
	}
};

/** メッセージの最大行数 */
TesSplitter.prototype.getMessageLineSize = function() {
	return this._messageLineSize;
};

/**
 * データを追加する
 */
TesSplitter.prototype.addData = function(tag, args) {
	this._data.push(this.createData(tag, args));
};

/**
 * データを作成する
 */
TesSplitter.prototype.createData = function(tag, args) {
	if (args == null) args = {};

	return {
		tag: tag,
		arguments: args,
		lineNumber: this.getLineNumber()
	};
};

/**
 * 分割処理（ラベル）
 */
TesSplitter.prototype.splitLabel = function(text) {
	var label = RubyCompatibility.stripText(text.substring(1));

	if (label == '') {
		throw new TesSyntaxError(
			'ラベルが無名です',
			'splitter.label.empty.error'
		);
	}

	this.addData('label', {'value': label});
};

/**
 * 分割処理（一般タグ）
 */
TesSplitter.prototype.splitNormal = function(text, tag) {
	this.addData(tag, this.getCommandArgs(text));
};

/** 行数取得 */
TesSplitter.prototype.getLineNumber = function() {
	if (this.eoc()) return -1;

	return this._lines[this._index].lineNumber;
};

/** コマンドの終端か？ */
TesSplitter.prototype.eoc = function() {
	return this._index >= this._lines.length;
};

/**
 * カレント行文字列取得
 */
TesSplitter.prototype.getLine = function(index) {
	if (index == null) index = this._index;
	if (index >= this._lines.length) return '';

	var text = this._lines[index].line;
	if (!text) return '';

	// 前後の空白を除去
	text = RubyCompatibility.stripText(text);

	// コメントを削除して返す
	switch (text[0]) {
		case '@':
			// タグ構文
			return this.eraceCommentNormal(text);
		case '*':
			// ラベル
			// ラベルも文字列リテラルの一種とみなす
		default:
			// 非タグ構文
			return this.eraceCommentText(text);
	}
};

/**
 * コメントの削除(タグ構文)
 */
TesSplitter.prototype.eraceCommentNormal = function(text) {
	var buffer = '';

	var quote = null;
	for (var i = 0; i < text.length; i++) {
		switch (text[i]) {
			case '"':
			case "'":
				buffer += text[i];
				// 引用符開始
				if (quote == text[i]) {
					quote = null;
				} else {
					quote = text[i];
				}

				continue;
			case '#':
				// 文字列中でなければ終了
				if (quote == null) return buffer;
			default:
				buffer += text[i];
		}
	}

	return buffer;
};

/**
 * コメントの削除(文字リテラル)
 */
TesSplitter.prototype.eraceCommentText = function(text) {
	var buffer = '';

	for (var i = 0; i < text.length; i++) {
		switch (text[i]) {
			case '#':
				return buffer;
			case '\\':
				if (text[i + 1] == '#') {
					buffer += text[++i];
					continue;
				}
			default:
				buffer += text[i];
		}
	}

	return buffer;
};

/**
 * メッセージのアンダースコア除去処理
 */
TesSplitter.prototype.replaceMessageUnderscore = function(text) {
	var buffer = '';
	for (var i = 0; i < text.length; i++) {
		switch (text[i]) {
			case '_':
				i++;
				continue;
			case '\\':
				if (text[i + 1] == '_') {
					buffer += '\\_';
					i += 2;
					break;
				}
			default:
				buffer += text[i];
				break;
		}
	}

	return buffer;
}

/**
 * 次の行文字列取得
 */
TesSplitter.prototype.getNextLine = function() {
	return this.getLine(this._index + 1);
};

/**
 * 一文字目取得
 */
TesSplitter.prototype.getFirstCharacter = function(text) {
	return RubyCompatibility.lstripText(text)[0];
};

/**
 * コマンド名取得
 */
TesSplitter.prototype.getCommandName = function(text) {
	if (this.getFirstCharacter(text) == '@') {
		return RubyCompatibility.stripText(text.match(/^\S+/)[0].substring(1)).toLowerCase();
	} else {
		return null;
	}
};

/**
 * コマンドの引数部を取得
 */
TesSplitter.prototype.getCommandArgs = function(text, arrayArgs) {
	var temp = {};
	var cmd = this.getCommandName(text);

	if (cmd) {
		var args = RubyCompatibility.stripText(text.substring(cmd.length + 1));
		if (args) {
			// 空白文字以外を検索
			var indexOfNotSpace = function(src, start) {
				if (start < 0) return -1;

				for (var i = start; i < src.length; i++) {
					switch (src[i]) {
						case '\r': case '\n': case ' ': case '\t':
							break;
						default:
							return i;
					}
				}

				return -1;
			};

			// 属性名の終端を検索
			var indexOfAttrNameEnd = function(src, start) {
				if (start < 0) return -1;

				for (var i = start; i < src.length; i++) {
					switch (src[i]) {
						case '\r': case '\n': case ' ': case '\t': case '=':
							return i;
					}
				}

				return -1;
			}

			// 値の開始位置を検索
			var indexOfquote = function(src, start) {
				if (start < 0) return -1;

				for (var i = start; i < src.length; i++) {
					switch (src[i]) {
						case '\r': case '\n': case ' ': case '\t':
							break;
						case '"': case "'":
							return i;
						default:
							return i;
					}
				}

				return -1;
			};

			var ptr = 0;
			while (ptr < args.length) {
				// 文字が発生するまでポインタを進める
				ptr = indexOfNotSpace(args, ptr);

				// 要素がなければ終了
				if (ptr < 0) break;

				// 要素名の最後まで走査
				var nextPtr = indexOfAttrNameEnd(args, ptr);

				var paramName;
				if (nextPtr >= 0) {
					paramName = args.substring(ptr, nextPtr);
				} else {
					throw new TesSyntaxError(
						'パラメータ値が見つかりません。',
						'syntax.params.value.notFound.error'
					);
				}

				// イコールを探す
				ptr = indexOfNotSpace(args, nextPtr);
				if (args[ptr] != '=') ptr = -1;

				// イコールが見つからない
				if (ptr < 0) {
					throw new TesSyntaxError(
						'パラメータ値が見つかりません。',
						'syntax.params.value.notFound.error'
					);
				}

				ptr++;
				// 値の開始位置を探す
				ptr = indexOfquote(args, ptr);

				// 値の開始位置が存在しない
				if (ptr < 0) {
					throw new TesSyntaxError(
						'パラメータ値が見つかりません。',
						'syntax.params.value.notFound.error'
					);
				}

				// 値のクォート文字を取得
				var quote;
				if (args[ptr] == '"' || args[ptr] == "'") {
					quote = args[ptr];
					ptr++;
				} else {
					quote = ' ';
				}

				// パラメータ値の取得
				var buffer = '';
				var closed = false;
				while (ptr < args.length && !closed) {
					if (args[ptr] == '\\') {
						// エスケープシーケンスの処理

						// エスケープ文字の次の文字が存在しない
						if (ptr + 1 >= args.length) {
							throw new TesSyntaxError(
								'エスケープシーケンスの指定が不正です。',
								'syntax.params.escape.invalid.error'
							);
						}

						var checkPart;
						if (ptr + 3 >= args.length) {
							checkPart = args.substr(ptr + 1, 3);

							if (/^[x]\d{2}$/.test(checkPart)) {
								// 8進数文字コード参照
								throw new TesSyntaxError(
									'非対応のエスケープシーケンスです。',
									'syntax.params.escape.notImplements.error',
									['\\' + checkPart]
								);
							} else if (/^\d{3}$/.test(checkPart)) {
								// 10進数文字コード参照
								throw new TesSyntaxError(
									'非対応のエスケープシーケンスです。',
									'syntax.params.escape.notImplements.error',
									['\\' + checkPart]
								);
							}

							//ptr += 4;
							//continue;
						}

						if (ptr + 5 >= args.length) {
							checkPart = args.substr(ptr + 1, 5);

							if (/^[u]\d{4}$/.test(checkPart)) {
								// Unicode文字コード参照
								throw new TesSyntaxError(
									'非対応のエスケープシーケンスです。',
									'syntax.params.escape.notImplements.error',
									['\\' + checkPart]
								);
							}

							//ptr += 6;
							//continue;
						}

						checkPart = args[ptr + 1];
						switch (checkPart) {
							case 'n':
								buffer += '\n';

								break;
							case 'f':
								buffer += '\f';

								break;
							case 'b':
								buffer += '\b';

								break;
							case 'r':
								buffer += '\r';

								break;
							case 't':
								buffer += '\t';

								break;
							case '\\':
								buffer += '\\';

								break;
							case "'":
								buffer += "'";

								break;
							case '"':
								buffer += '"';

								break;
							default:
								throw new TesSyntaxError(
									'非対応のエスケープシーケンスです。',
									'syntax.params.escape.notImplements.error',
									['\\' + checkPart]
								);
						}
					} else if (quote == ' ' && '\r\n \t'.indexOf(args[ptr]) >= 0) {
						ptr--;
						closed = true;
					} else if (quote != ' ' && args[ptr] == quote) {
						closed = true;
					} else {
						buffer += args[ptr];
					}

					ptr++;
				}

				if (quote != ' ' && !closed) {
					throw new TesSyntaxError(
						'パラメータ値の終端記号が見つかりません。',
						'syntax.params.quote.notFound.error'
					);
				}
				temp[paramName] = buffer;

				if (arrayArgs != null) {
					if (arrayArgs[paramName] == null) {
						arrayArgs[paramName] = [];
					}

					arrayArgs[paramName].push(buffer);
				}
			}
		}
	}

	return temp;
};

/**
 * 次の行がメッセージか？
 */
TesSplitter.prototype.nextMessage = function() {
	var text = this.getNextLine();

	// 空行はfalse
	if (RubyCompatibility.isEmptyText(text)) return false;

	// 特殊文字行がきたらfalse
	var first = this.getFirstCharacter(text);
	return (!(first == '*' || first == '@'))
};

/**
 * メッセージの置換処理
 */
TesSplitter.prototype.replaceMessage = function(text) {
	var defaultMessageReplacers = {
		'\\\\#':'#',
		'\\\\@':'@',
		'\\\\\\*':'*',
		'\\\\_':'_'
	};

	for (var before in defaultMessageReplacers) {
		var after = defaultMessageReplacers[before];

		text = text.replace(new RegExp(before, 'g'), after);
	}

	for (var before in this._replacers) {
		var after = this._replacers[before];

		text = text.replace(new RegExp(before, 'g'), after);
	}

	return text;
};

/**
 * ＴＥＳバリデータクラス
 */
var TesValidator = function(engine) {
	this.initialize.apply(this, arguments);
};
TesValidator.prototype = Object.create(TesConverterBase.prototype);
TesValidator.prototype.constructor = TesValidator;

/**
 * 初期化
 */
TesValidator.prototype.initialize = function(engine) {
	TesConverterBase.prototype.initialize.call(this, engine);

	this._validates = {};
	this._validateFuncs = {};
	this.addBasicValidateFuncs();
};

/**
 * リセット処理
 */
TesValidator.prototype.reset = function() {
	TesConverterBase.prototype.reset.call(this);

	this._index = 0;
};

/**
 * 処理開始
 */
TesValidator.prototype.run = function(data) {
	this._data = data;

	this.engine.writeMessage('>TESのコマンド検証を開始', 'validator.start');

	try {
		TesConverterBase.prototype.run.call(this);
	} catch (err) {
		this.writeError(err);
		this.engine.writeErrorMessage('>>TESのコマンド検証に失敗', 'validator.failure');
	}

	this.engine.writeErrorMessage('>>>TESのコマンド検証処理が完了\n', 'validator.complete');
};

/**
 * メイン処理
 */
TesValidator.prototype.processing = function() {
	while (!this.eoc()) {
		try {
			this.validate();
		} catch (err) {
			if (err.name == 'ValidateError') {
				this.writeError(err);
			} else {
				throw err;
			}
		} finally {
			this._index++;
		}
	}
};

/**
 * 後処理
 */
TesValidator.prototype.postProcessing = function() {
	TesConverterBase.prototype.postProcessing.call(this);

	if (!this.isError() && this._data.length == 0) {
		throw new TesValidateError(
			'\n>>TESコマンドの検証で問題が見つかりました',
			'validator.nodata.error'
		);
	}
};

/**
 * コマンドの検証
 */
TesValidator.prototype.validate = function() {
	var data = this._data[this._index];

	var validates = this.getValidates(data.tag);
	for (var paramName in validates) {
		var funcs = validates[paramName];

		for (var func in funcs) {
			var vargs = funcs[func];

			try {
				if (this._validateFuncs[func] == null) {
					throw new TesValidateError(
						'検証コマンド「' + func + '」がありません',
						'validator.function.notFound.error',
						[func]
					);
				}

				this._validateFuncs[func].call(this, vargs, paramName, data.arguments);
			} catch (err) {
				if (err.name == 'ValidateError') {
					this.writeError(err);
				} else {
					throw err;
				}
			}
		}
	}
};

/**
 * バリデーションの取得
 */
TesValidator.prototype.getValidates = function(tag) {
	var validate = this._validates[tag];

	if (validate) {
		return validate;
	} else {
		return {};
	}
};

/** 行数取得 */
TesValidator.prototype.getLineNumber = function() {
	if (this._data[this._index] == null) return -1;

	return this._data[this._index].lineNumber;
};

/** コマンドの終端か？ */
TesValidator.prototype.eoc = function() {
	return this._index >= this._data.length;
};


/**
 * コマンド検証関数をセット
 */
TesValidator.prototype.addValidateFuncs = function(validateFuncs) {
	for (var key in validateFuncs) {
		this._validateFuncs[key] = validateFuncs[key];
	}
};

/** 基本のバリデーション関数を追加する */
TesValidator.prototype.addBasicValidateFuncs = function() {
	this.addValidateFuncs(BasicValidateFunctions);
};

/** パラメータ値の解決を行う */
TesValidator.prototype.resolveParameter = function(value, params) {
	CommonFunctions.resolveParameter(value, params, ValidateError);
};

/**
 * ＴＥＳコンバータクラス
 */
var TesConverter = function(engine) {
	this.initialize.apply(this, arguments);
};
TesConverter.prototype = Object.create(TesConverterBase.prototype);
TesConverter.prototype.constructor = TesConverter;

/**
 * 初期化
 */
TesConverter.prototype.initialize = function(engine) {
	TesConverterBase.prototype.initialize.call(this, engine);

	this._convertFuncs = {};
};

/**
 * リセット処理
 */
TesConverter.prototype.reset = function() {
	TesConverterBase.prototype.reset.call(this);

	this._index = 0;
	this._indent = 0;
};

/**
 * 処理開始
 */
TesConverter.prototype.run = function(commands) {
	this._commands = commands;
	this._data = [];

	this.engine.writeMessage('>イベントデータへの変換を開始', 'converter.start');

	try {
		TesConverterBase.prototype.run.call(this);
	} catch (err) {
		this.writeError(err);
		this.engine.writeErrorMessage('>>イベントデータへの変換に失敗', 'converter.failure');
	}

	this.engine.writeMessage('>>>イベントデータへの変換処理が完了\n', 'converter.complete');
};

/**
 * メイン処理
 */
TesConverter.prototype.processing = function() {
	while (!this.eoc()) {
		try {
			this.convert();
		} catch (err) {
			if (err.name == 'ConvertError') {
				this.writeError(err);
			} else {
				throw err;
			}
		} finally {
			this._index++;
		}
	}
};

/**
 * 後処理
 */
TesConverter.prototype.postProcessing = function() {
	TesConverterBase.prototype.postProcessing.call(this);

	if (!this.isError() && this._data.length == 0) {
		throw new TesConvertError(
			'\n>>変換されたイベントがありませんでした',
			'converter.nodata.error'
		);
	}
};

/**
 * コマンドの変換
 */
TesConverter.prototype.convert = function() {
	var command = this._commands[this._index];

	if (this._convertFuncs[command.tag] == null) {
		throw new TesConvertError(
			'変換コマンド「' + command.tag + '」がありません',
			'converter.function.notFound.error',
			[command.tag]
		);
	}

	var data = this._convertFuncs[command.tag].call(this, command.arguments);

	if (data) this._data.push(data);
};

/**
 * リスト値の変換
 */
TesConverter.prototype.convertList = function(value, list, defaultValue) {
	var listValue = list[value];

	if (listValue === undefined) {
		listValue = defaultValue;
	}

	return listValue;
};

/**
 * フラグの変換処理
 */
TesConverter.prototype.convertFlag = function(flag, defaultValue) {
	return this.convertList(flag, {'on': 0, 'off': 1}, 0);
};

/** 行数取得 */
TesConverter.prototype.getLineNumber = function() {
	return this._commands[this._index].lineNumber;
};

/**
 * コマンドの終端か？
 */
TesConverter.prototype.eoc = function() {
	return this._index >= this._commands.length;
};

var TesDeconverter = function(engine) {
	this.initialize.apply(this, arguments);
};
TesDeconverter.prototype = Object.create(TesConverterBase.prototype);
TesDeconverter.prototype.constructor = TesDeconverter;

/**
 * 初期化
 */
TesDeconverter.prototype.initialize = function(engine) {
	TesConverterBase.prototype.initialize.call(this, engine);

	this._deconvertFuncs = {};
	this._targetIdentifiers = [];
	this.quote = '"';
};

/**
 * リセット処理
 */
TesDeconverter.prototype.reset = function() {
	TesConverterBase.prototype.reset.call(this);

	this._index = 0;
};

/**
 * 処理開始
 */
TesDeconverter.prototype.run = function(target) {
	this._data = [];
	this._target = target;
	this._events = target.list;

	this.engine.writeMessage('>イベントデータからの逆変換を開始', 'deconverter.start', [], this.getMessageAddress(target));

	try {
		TesConverterBase.prototype.run.call(this);
	} catch (err) {
		this.writeError(err);
		this.engine.writeErrorMessage('>>イベントデータからの逆変換に失敗', 'deconverter.failure', [], this.getMessageAddress(target));
	}

	this.engine.writeMessage('>>>イベントデータからの逆変換処理が完了\n', 'deconverter.complete', [], this.getMessageAddress(target));
};

/**
 * メッセージアドレスを取得する。
 */
TesDeconverter.prototype.getMessageAddress = function(target, lineNumber, code) {
	if (target != null && target.attrs != null) {
		var attrs = {};

		for (var key in target.attrs) {
			attrs[key] = target.attrs[key];
		}

		if (lineNumber != null) attrs.lineNumber = lineNumber;
		if (code != null) attrs.code = code;

		var address = '';
		for (var key in attrs) {
			if (address) address += ', ';

			address += key + ':';
			if (attrs[key] != null) {
				address += attrs[key].toString();
			}
		}

		if (address != '') return address;
	}

	if (lineNumber != null) {
		return lineNumber.toString();
	} else {
		return null;
	}
};

/**
 * メイン処理
 */
TesDeconverter.prototype.processing = function() {
	while (!this.eoc()) {
		try {
			this.deconvert();
		} catch (err) {
			if (err.name == 'DeconvertError') {
				this.writeError(err);
			} else {
				throw err;
			}
		} finally {
			this._index++;
		}
	}
};

/**
 * 後処理
 */
TesDeconverter.prototype.postProcessing = function() {
	TesConverterBase.prototype.postProcessing.call(this);

	if (!this.isError() && this._data.length == 0) {
		throw new TesConvertError(
			'\n>>逆変換されたイベントがありませんでした',
			'deconverter.nodata.error'
		);
	}
};

/**
 * コマンドの変換
 */
TesDeconverter.prototype.deconvert = function() {
	var event = this._events[this._index];

	if (this._deconvertFuncs[event.code] == null) {
		throw new TesDeconvertError(
			'逆変換コマンド「' + event.code + '」がありません',
			'deconverter.function.notFound.error',
			[event.code]
		);
	}

	this._code = event.code;
	var data = this._deconvertFuncs[event.code].call(this, event);

	if (data != null) {
		if (data.indent == null && typeof(data) != 'string' && event.indent != null) {
			data.indent = event.indent;
		}

		this._data.push(data);
	}
};

/** 行数取得 */
TesDeconverter.prototype.getLineNumber = function() {
	return this._index;
};

/**
 * コマンドの終端か？
 */
TesDeconverter.prototype.eoc = function() {
	return this._index >= this._events.length;
};

/**
 * デコンパイル対象の判定
 */
TesDeconverter.prototype.identifyTargets = function(basename, data) {
	for (var i = 0; i < this._targetIdentifiers.length; i++) {
		var targetIdentifier = this._targetIdentifiers[i];
		var targets = targetIdentifier.call(this, basename, data);
		
		if (targets != null) return targets;
	}

	if (data != null && Array.isArray(data)) {
		// データが配列ならデフォルトターゲット
		return {
			type: 'tes-event',
			targets: [{
				name: basename,
				list: data,
				attrs: {}
			}]
		};
	}

	throw new TesDeconvertError(
		'この形式のデータをデコンパイルすることはできません。',
		'deconverter.target.notIdentified.error'
	);
};

/**
 * 値の逆変換
 */
TesDeconverter.prototype.deconvertValue = function(argName, value, defaultValue) {
	if (value === undefined) {
		if (defaultValue === undefined) {
			throw new Tes.DeconvertError(
				'[' + argName + '] : 不正な値が指定されました。',
				'deconverter.undefined.error',
				[argName]
			);
		} else {
			return null;
		}
	}

	if (defaultValue !== undefined) {
		if (Array.isArray(value)) {
			if (value.toString() == defaultValue.toString()) return null;
		} else {
			if (value === defaultValue) return null;
		}
	}

	return value;
};

/**
 * 文字列値の逆変換
 */
TesDeconverter.prototype.deconvertString = function(argName, args, value, defaultValue) {
	value = this.deconvertValue(argName, value, defaultValue);
	if (value == null) return;

	if (typeof(value) != 'string') {
		throw new Tes.DeconvertError(
			'[' + argName + '] : 不正な値が指定されました。',
			'deconverter.undefined.error',
			[argName]
		);
	}

	args[argName] = value;
};

/**
 * 数値の逆変換
 */
TesDeconverter.prototype.deconvertNumeric = function(argName, args, value, defaultValue) {
	value = this.deconvertValue(argName, value, defaultValue);
	if (value == null) return;

	if (typeof(value) != 'number') {
		throw new Tes.DeconvertError(
			'[' + argName + '] : 不正な値が指定されました。',
			'deconverter.undefined.error',
			[argName]
		);
	}

	args[argName] = value;
};

/**
 * 論理値の逆変換
 */
TesDeconverter.prototype.deconvertBoolean = function(argName, args, value, defaultValue) {
	value = this.deconvertValue(argName, value, defaultValue);
	if (value == null) return;

	if (typeof(value) != 'boolean') {
		throw new Tes.DeconvertError(
			'[' + argName + '] : 不正な値が指定されました。',
			'deconverter.undefined.error',
			[argName]
		);
	}

	args[argName] = value;
};

/**
 * 配列値の逆変換
 */
TesDeconverter.prototype.deconvertArray = function(argName, args, value, defaultValue) {
	value = this.deconvertValue(argName, value, defaultValue);
	if (value == null) return;

	if (!Array.isArray(value)) {
		throw new Tes.DeconvertError(
			'[' + argName + '] : 不正な値が指定されました。',
			'deconverter.undefined.error',
			[argName]
		);
	}

	args[argName] = value;
};

/**
 * リスト値の逆変換
 */
TesDeconverter.prototype.deconvertList = function(argName, args, value, list, defaultValue) {
	var orgValue = this.deconvertValue(argName, value, defaultValue);
	if (orgValue == null) return;

	var listValue = undefined;
	for (var i = 0; i < list.length; i += 2) {
		if (list[i] == value) {
			listValue = list[i + 1];
			break;
		}
	}

	if (listValue === undefined) {
		if (defaultValue === undefined) {
			throw new Tes.DeconvertError(
				'[' + argName + '] : 不正な値が指定されました。',
				'deconverter.undefined.error',
				[argName]
			);
		} else {
			listValue = defaultValue;
		}
	}

	args[argName] = listValue;
};

/**
 * フラグ値の逆変換
 */
TesDeconverter.prototype.deconvertFlag = function(argName, args, value, defaultValue) {
	this.deconvertList(
		argName,
		args, 
		value, [
			0, 'on',
			1, 'off'
		], 
		defaultValue
	);
};

/**
 * メッセージの逆変換
 */
TesDeconverter.prototype.deconvertMessage = function(argName, value) {
	var defaultMessageReplacers = {
		'#':'\\#',
		'@':'\\@',
		'\\*':'\\*',
		'_':'\\_'
	};

	value = this.deconvertValue(argName, value);
	if (value == null) return '';

	for (var before in defaultMessageReplacers) {
		var after = defaultMessageReplacers[before];

		value = value.replace(new RegExp(before, 'g'), after);
	}

	return value;
}

/**
 * オーディオオブジェクトの逆変換
 */
TesDeconverter.prototype.deconvertAudio = function(nameName, volumeName, pitchName, panName, args, audio, defaultName) {
	if (audio == null) {
		throw new Tes.DeconvertError(
			'オーディオオブジェクトが未設定です。',
			'deconverter.audio.undefined.error'
		);
	}

	this.deconvertString(nameName, args, audio.name, defaultName);
	this.deconvertNumeric(volumeName, args, audio.volume);
	this.deconvertNumeric(pitchName, args, audio.pitch);
	this.deconvertNumeric(panName, args, audio.pan);
};

/**
 * カラーオブジェクトの逆変換
 */
TesDeconverter.prototype.deconvertColor = function(redName, greenName, blueName, grayName, args, color) {
	if (color == null) {
		throw new Tes.DeconvertError(
			'カラーオブジェクトが未設定です。',
			'deconverter.color.undefined.error'
		);
	}

	if (redName != null) this.deconvertNumeric(redName, args, color[0]);
	if (greenName != null) this.deconvertNumeric(greenName, args, color[1]);
	if (blueName != null) this.deconvertNumeric(blueName, args, color[2]);
	if (grayName != null) this.deconvertNumeric(grayName, args, color[3]);
}

/**
 * トーンオブジェクトの逆変換
 */
TesDeconverter.prototype.deconvertTone = function(redName, greenName, blueName, strengthName, args, tone) {
	if (tone == null) {
		throw new Tes.DeconvertError(
			'トーンオブジェクトが未設定です。',
			'deconverter.tone.undefined.error'
		);
	}

	if (redName != null) this.deconvertNumeric(redName, args, tone[0]);
	if (greenName != null) this.deconvertNumeric(greenName, args, tone[1]);
	if (blueName != null) this.deconvertNumeric(blueName, args, tone[2]);
	if (strengthName != null) this.deconvertNumeric(strengthName, args, tone[3]);
}

/** エラー出力 */
TesDeconverter.prototype.writeError = function(err) {
	this._error = err;

	this.engine.writeError(err, this.getMessageAddress(this._target, this.getLineNumber(), this._code));
};

/**
 * ソースコードを取得する。
 */
TesDeconverter.prototype.getSource = function() {
	var datas = this.getData();
	var source = '';

	for (var i = 0; i < datas.length; i++) {
		source += this.getSourceRecursive(datas[i]);
	}

	return source;
};

/**
 * ソースコードを取得する再帰処理。
 */
TesDeconverter.prototype.getSourceRecursive = function(data) {
	var source = '';

	if (typeof(data) == 'string') {
		return data + '\n';
	} else if (Array.isArray(data)) {
		for (var i = 0; i < data.length; i++) {
			source += this.getSourceRecursive(data[i]);
		}

		return source;
	}

	for (var j = 0; j < data.indent; j++) {
		source += '  ';
	}

	var putArgs = (name, value) => {
		if (Array.isArray(value)) {
			for (var k = 0; k < value.length; k++) {
				putArgs(name, value[k]);
			}
		} else {
			source += ' ' + name + '=' + this.quote + serializeValue(value) + this.quote;
		}
	};

	var serializeValue = (value) => {
		if (typeof(value) != 'string') return value.toString();

		var buffer = '';

		for (var i = 0; i < value.length; i++) {
			switch (value[i]) {
				case '\n':
					buffer += '\\n'; break;
				case '\f':
					buffer += '\\f'; break;
				case '\b':
					buffer += '\\b'; break;
				case '\r':
					buffer += '\\r'; break;
				case '\t':
					buffer += '\\t'; break;
				case '\\':
					buffer += '\\\\'; break;
				case '"':
					buffer += '\\"'; break;
				case "'":
					buffer += "\\'"; break;
				default:
					buffer += value[i]; break;
			}
		}

		return buffer;
	};

	source += '@' + data.name;
	for (var key in data.args) {
		var arg = data.args[key];

		putArgs(key, arg);
	}

	return source + '\n';
};

var TesLinker = function(engine) {
	this.initialize.apply(this, arguments);
};
TesLinker.prototype = Object.create(TesConverterBase.prototype);
TesLinker.prototype.constructor = TesLinker;

/**
 * 初期化
 */
TesLinker.prototype.initialize = function(engine) {
	TesConverterBase.prototype.initialize.call(this, engine);

	this._linkFuncs = {};
};

/**
 * リセット処理
 */
TesLinker.prototype.reset = function() {
	TesConverterBase.prototype.reset.call(this);
};

/**
 * 処理開始
 */
TesLinker.prototype.run = function(source) {
	this._linkType = source.linkType;
	this._target = source.target;
	this._targetData = source.targetData;
	this._eventList = source.eventList;
	this._lineNumber = source.lineNumber;
	this._data = [];

	this.engine.writeMessage('>イベントデータのリンクを開始', 'linker.start', []);

	try {
		TesConverterBase.prototype.run.call(this);
	} catch (err) {
		this.writeError(err);
		this.engine.writeErrorMessage('>>イベントデータのリンクに失敗', 'linker.failure', []);
	}

	this.engine.writeMessage('>>>イベントデータのリンク処理が完了\n', 'linker.complete', []);
};

/**
 * メイン処理
 */
TesLinker.prototype.processing = function() {
	TesConverterBase.prototype.processing.call(this);

	try {
		if (!this._linkFuncs[this._linkType] == null) {
			throw new TesDeconvertError(
				'リンク関数「' + this._linkType + '」がありません',
				'linker.function.notFound.error',
				[this._linkType]
			);
		}

		var linkFunc = this._linkFuncs[this._linkType];
		linkFunc.call(
			this,
			this._target,
			this._targetData,
			this._eventList
		);
	} catch (err) {
		if (err.name == 'LinkError') {
			this.writeError(err, this.getLineNumber());
		} else {
			throw err;
		}
	}
};

/** 行数取得 */
TesLinker.prototype.getLineNumber = function() {
	return this._lineNumber;
};

/**
 * コマンドの終端か？
 */
TesLinker.prototype.eoc = function() {
	return true;
};

/**
 * Ruby互換関数クラス
 */
var RubyCompatibility = {};

/**
 * 左側の空白文字を取り除く
 */
RubyCompatibility.lstripText = function(text) {
	if (text == null) return text;

	return text.replace(/^[ \t\r\n\f\v]*/g, '');
};
/**
 * 右側の空白文字を取り除く
 */
RubyCompatibility.rstripText = function(text) {
	if (text == null) return text;

	return text.replace(/[ \t\r\n\f\v\0]*$/g, '');
};
/**
 * 空白文字を取り除く
 */
RubyCompatibility.stripText = function(text) {
	return RubyCompatibility.lstripText(RubyCompatibility.rstripText(text));
};
/**
 * 空白文字か？
 */
RubyCompatibility.isEmptyText = function(text) {
	if (text == null) return true;

	return RubyCompatibility.stripText(text) == '';
};

/**
 * 汎用関数
 */
var CommonFunctions = {};

 /**
  * 数値チェック
  */
CommonFunctions.checkNumeric = function(paramName, params, amin, amax, errorClass) {
	if (errorClass == null) errorClass = Error;

	var param = params[paramName];
	if (param == null) return param;

	// 型判定
	if (!/^[-+0-9]\d*$/.test(param)) {
		throw new errorClass(
			'[' + paramName + '] : 数値以外が代入されました。',
			'validator.numeric.invalid.error',
			[paramName]
		)
	}

	var paramValue = parseInt(param, 10);

	// 最小値判定
	if (amin) {
		var min = CommonFunctions.resolveParameter(amin, params, errorClass);

		if (paramValue < min) {
			throw new errorClass(
				'[' + paramName + '] : 最小値より小さい値が使われています。',
				'validator.numeric.min.error',
				[paramName]
			);
		}
	}

	// 最大値判定
	if (amax) {
		var max = CommonFunctions.resolveParameter(amax, params, errorClass);

		if (paramValue > max) {
			throw new errorClass(
				'[' + paramName + '] : 最大値より大きい値が使われています。',
				'validator.numeric.max.error',
				[paramName]
			);
		}
	}

	return paramValue;
};

/**
 * 必須チェック
 */
CommonFunctions.checkNotEmpty = function(paramName, params, errorClass) {
	if (errorClass == null) errorClass = Error;

	if (params[paramName] == null || RubyCompatibility.isEmptyText(params[paramName])) {
		throw new errorClass(
			'[' + paramName + '] : 必須引数が省略されています。',
			'default.notEmpty.error',
			[paramName]
		);
	}
};

/**
 * 真偽値チェック
 */
CommonFunctions.checkBool = function(paramName, params, errorClass) {
	if (errorClass == null) errorClass = Error;

	if (params[paramName] == null) return;
	if (!/^(true|false)$/i.test(params[paramName])) {
		throw new errorClass(
			'[' + paramName + '] : true/false 以外が代入されました。',
			'validator.isBool.error',
			[paramName]
		);
	}
};

/**
 * リストチェック
 */
CommonFunctions.checkList = function(paramName, params, list, errorClass) {
	if (errorClass == null) errorClass = Error;

	if (params[paramName] == null) return;
	if (list.indexOf(params[paramName]) < 0) {
		throw new errorClass(
			'[' + paramName + '] : 許可されていない値が使われました。',
			'validator.list.error',
			[paramName]
		);
	}
};

/**
 * 0チェック
 */
CommonFunctions.checkZero = function(paramName, params, errorClass) {
	if (errorClass == null) errorClass = Error;

	if (params[paramName] == null) return;
	if (parseInt(params[paramName], 10) == 0) {
		throw new errorClass(
			'[' + paramName + '] : 「0」が使われました。',
			'validator.zeroCheck.error',
			[paramName]
		);
	}
};

/**
 * 正規表現チェック
 */
CommonFunctions.checkRegex = function(paramName, params, regex, errorClass) {
	if (errorClass == null) errorClass = Error;

	if (params[paramName] == null) return;
	if (!regex.test(params[paramName])) {
		throw new errorClass(
			'[' + paramName + '] : 正規表現にマッチしませんでした。',
			'validator.regCheck.error',
			[paramName]
		);
	}
}

/** パラメータ値の解決を行う */
CommonFunctions.resolveParameter = function(value, params, errorClass) {
	if (errorClass == null) errorClass = Error;

	if (value == null) return value;
	if (typeof value == 'string') {
		if (value[0] == ':') {
			var symbolValue = params[value.substring(1)];
			if (symbolValue === undefined) {
				throw new errorClass(
					'シンボルの解決に失敗しました。',
					'default.symbol.notFound.error',
					[value.substring(1)]
				);
			}

			return symbolValue;
		} else {
			return eval(value);
		}
	} else {
		return parseInt(value, 10);
	}
};

Tes.CommonFunctions = CommonFunctions;

/**
 * 基本のバリデーション関数
 */
var BasicValidateFunctions = {};

// 必須要素チェック
BasicValidateFunctions.notEmpty = function(vargs, paramName, params) {
	CommonFunctions.checkNotEmpty(paramName, params, TesValidateError);
};

// 真偽値チェック
BasicValidateFunctions.isBool = function(vargs, paramName, params) {
	CommonFunctions.checkBool(paramName, params, TesValidateError);
};

// 数値チェック
BasicValidateFunctions.isNumeric = function(vargs, paramName, params) {
	CommonFunctions.checkNumeric(
		paramName,
		params,
		vargs[0],
		vargs[1],
		TesValidateError
	);
};

// リストチェック
BasicValidateFunctions.list = function(vargs, paramName, params) {
	CommonFunctions.checkList(
		paramName,
		params,
		vargs,
		TesValidateError
	);
};

// 0チェック
BasicValidateFunctions.zeroCheck = function(vargs, paramName, params) {
	CommonFunctions.checkZero(paramName, params, TesValidateError);
};

// 正規表現チェック
BasicValidateFunctions.regCheck = function(vargs, paramName, params) {
	CommonFunctions.checkRegex(
		paramName,
		params,
		vargs[0],
		TesValidateError
	);
};

Tes.BasicValidateFunctions = BasicValidateFunctions;

/**
 * 文字列リソース
 */
var StringResourceData = {};

StringResourceData.default = {
	"default": {
		"notEmpty.error": "{0} : 必須引数が省略されています。",
		"symbol.notFound.error":"シンボルが解決できません。シンボル: {0}"
	},
	"loader": {
		"start": "TESの読み込みを開始",
		"failure": "TESの読み込みに失敗",
		"complete": "TESの読み込み処理が完了",
		"nodata.error": "変換できるシナリオがありませんでした"
	},
	"splitter": {
		"start": "TESのコマンド分割を開始",
		"failure": "TESのコマンド分割に失敗",
		"complete": "TESのコマンド分割処理が完了",
		"nodata.error": "TESコマンドが検出されませんでした",
		"label.empty.error": "ラベルが無名です"
	},
	"validator": {
		"start": "TESのコマンド検証を開始",
		"failure": "TESのコマンド検証に失敗",
		"complete": "TESのコマンド検証処理が完了",
		"nodata.error": "TESコマンドの検証対象が見つかりませんでした",
		"function.notFound.error": "検証メソッドが見つかりません。メソッド: {0}",
		"isBool.error": "{0} : true/false 以外が代入されました。",
		"list.error": "{0} : 許可されていない値が使われました。",
		"zeroCheck.error": "{0} : 0 を代入することはできません。",
		"regCheck.error": "{0} : 正規表現にマッチしませんでした。",
		"numeric": {
			"invalid.error": "{0} : 数値以外が代入されました。",
			"min.error": "{0} : 最小値より小さい値が使われています。",
			"max.error": "{0} : 最大値より大きい値が使われています。"
		}
	},
	"converter": {
		"start": "イベントデータへの変換を開始",
		"failure": "イベントデータへの変換に失敗",
		"complete": "イベントデータへの変換処理が完了",
		"nodata.error": "変換されたイベントがありませんでした",
		"function.notFound.error": "変換メソッドが見つかりません。メソッド: {0}"
	},
	"syntax": {
		"params": {
			'value.notFound.error': "パラメータ値が見つかりません。",
			"quote.notFound.error": "パラメータ値の終端記号が見つかりません。",
			"escape.invalid.error": "エスケープシーケンスの指定が不正です。",
			"escape.notImplements.error": "非対応のエスケープシーケンスです。文字: {0}"
		}
	},
	"deconverter": {
		"start": "イベントデータからの逆変換を開始",
		"failure": "イベントデータからの逆変換に失敗",
		"complete": "イベントデータからの逆変換処理が完了",
		"nodata.error": "逆変換されたイベントがありませんでした",
		"function.notFound.error": "逆変換メソッドが見つかりません。メソッド: {0}",
		"target.notIdentified.error": "この形式のデータをデコンパイルすることはできません。",
		"undefined.error": "{0} : 不正な値が指定されました。",
		"audio": {
			"undefined.error": "オーディオオブジェクトが未設定です。",
		},
		"color": {
			"undefined.error": "カラーオブジェクトが未設定です。",
		},
		"tone": {
			"undefined.error": "トーンオブジェクトが未設定です。",
		}
	},
	"linker": {
		"start": "イベントデータのリンクを開始",
		"failure": "イベントデータのリンクに失敗",
		"complete": "イベントデータのリンク処理が完了",
		"function.notfound.error": "リンク関数が見つかりません。メソッド: {0}"
	}
};

Tes.StringResourceData = StringResourceData;

return Tes;
});
