'use strict';

var fs = require('fs-extra');
var path = require('path');
var Tes = require('uotes-core');
var StringResourceManager = require('es-common-core/string-resource-manager');

/**
 * uoTES.js リンカーモジュール
 * リンカークラス
 * 
 * @since 2016/09/15
 * @author RaTTiE
 */
class Linker {
	/**
	 * コンストラクタ。
	 */
	constructor() {
		this.version = require('./package.json').version;
		this.stringResource = new StringResourceManager();
		this.engine = new Tes.Engine();
		this.console = console;
	}

	/**
	 * モジュールのルートパスを取得する。
	 */
	get moduleRootPath() {
		return path.resolve(__dirname, '.');
	}

	/**
	 * リソースパスを追加する。
	 */
	addResourcePath(resourcePath) {
		this.stringResource.addPath(resourcePath);
	}

	/** 言語を取得する。 */
	get language() {
		return this.stringResource.language;
	}

	/** 言語を設定する。 */
	set language(value) {
		this.stringResource.language = value;
	}

	/** メッセージライターを取得する。 */
	get messageWriter() {
		return this.engine.messageWriter;
	}

	/** メッセージライターを設定する。 */
	set messageWriter(value) {
		this.engine.messageWriter = value;
	}

	/**
	 * 文字列リソースのロードを行う。
	 */
	loadStringResource() {
		var that = this;
		var existsLanguage = this.stringResource.list().find((element, index, array) => {
			return element.language == that.stringResource.language;
		}) != null;
		if (!existsLanguage) this.stringResource.language = 'default';
		
		this.stringResource.loadFromObject(Tes.StringResourceData);
		this.stringResource.load(this.stringResource.language, true);
	}

	/**
	 * モジュールのロードを行う。
	 * 
	 * @param moduleNames ロード対象モジュール名の一覧。
	 */
	loadModules(moduleNames) {
		for (var i = 0; i < moduleNames.length; i++) {
			this.loadModule(moduleNames[i]);
		}
	}

	/**
	 * モジュールのロードを行う。
	 * 
	 * @param moduleName ロード対象モジュール名。
	 */
	loadModule(moduleName) {
		var module = require('uotes-module-' + moduleName);

		try {
			this.console.log(this.stringResource.get('module.loading', moduleName));
			this.engine.addModule(module);
			if (module.stringResourceData) {
				this.stringResource.loadFromObject(
					module.stringResourceData,
					this.stringResource.language,
					true
				);
			}
		} catch (err) {
			this.console.error(this.stringResource.get('module.load.failure', moduleName));
  
			throw err;
		}
	}

	/**
	 * リンク処理を行う。
	 * 
	 * @param inFile 入力ファイルのパス。
	 * @param outFile 出力ファイルのパス。
	 * @return 処理成功時はtrue。エラー時はfalse。
	 */
	execute(inFile, outFile) {
		var inPath = path.resolve('./', inFile);
		var outPath = path.resolve('./', outFile);

		try {
			// リンクファイルの取得
			var linkData;
			try {
				linkData = JSON.parse(fs.readFileSync(inPath, 'utf8'));
			} catch (err) {
				if (err.code == 'ENOENT') {
					this.console.error(this.stringResource.get('tes.link.source.notFound.error'));

					return false;
				} else {
					this.console.error(this.stringResource.get('tes.link.source.load.failure'));

					throw err;
				}
			}

			var linkType = linkData.type;
			var targets = linkData.targets;

			if (linkType == null || !Array.isArray(targets)) {
				this.console.error(this.stringResource.get('tes.link.source.load.failure'));

				return false;
			}

			// リンクタイプの存在判定は先に行う
			if (!this.engine.existsLinkFunction(linkType)) {
				this.console.error(this.stringResource.get('tes.link.type.notFound.error', linkType));

				return false;
			}

			// リンク対象ファイルのロード
			var targetData;
			try {
				targetData = JSON.parse(fs.readFileSync(outPath, 'utf8'));
			} catch (err) {
				if (err.code == 'ENOENT') {
					this.console.error(this.stringResource.get('tes.link.destination.notFound.error'));

					return false;
				} else {
					this.console.error(this.stringResource.get('tes.link.destination.load.failure'));

					throw err;
				}
			}

			// リンク処理実施
			var hasError = false;

			var inDir = path.dirname(inPath);
			for (var i = 0; i < targets.length; i++) {
				var target = targets[i];
				var eventFile = target.name + '.json';
				var eventPath = path.resolve(inDir, eventFile);

				var eventList;
				try {
					eventList = JSON.parse(fs.readFileSync(eventPath, 'utf8'));
				} catch (err) {
					if (err.code == 'ENOENT') {
						this.console.error(this.stringResource.get('tes.link.event.notFound.error'));

						return false;
					} else {
						this.console.error(this.stringResource.get('tes.link.event.load.failure'));

						throw err;
					}
				}

				if (!Array.isArray(eventList)) {
					this.console.error(this.stringResource.get('tes.link.event.load.failure'));

					throw err;
				}

				this.engine.messageWriter.fileName = eventFile;
				var result = this.engine.link(
					linkType,
					i,
					target,
					targetData,
					eventList
				);

				if (result.error != null) {
					hasError = true;
					continue;
				}
			}

			// リンクエラー時はファイル更新しない
			if (hasError) {
				this.console.error(this.stringResource.get('tes.link.write.ignore'));

				return;
			}

			// リンク対象ファイルの更新
			try {
				outFile = path.basename(outPath);

				this.console.log(this.stringResource.get('tes.link.writing', outFile));
				fs.writeFileSync(outPath, JSON.stringify(targetData));
			} catch (err) {
				this.console.error(this.stringResource.get('tes.link.write.failure'));

				throw err;
			}

			return true;
		} catch (err) {
			this.console.error(this.stringResource.get('tes.link.failure'));

			throw err;
		}
	}
}

module.exports = Linker;
